#! /usr/bin/perl -w

%non_aggregate = ( GdkAtom => 1, GdkDragProtocol => 1, GdkIMStyle => 1 );
%non_basename = ( foreground => 1, background => 1, name => 1, size => 1 );

%redeclare = ( "gtk_object" => "gtk_signal", "gdk_window" => "A9" );    
%redeclare_source = ( "gtk_signal" => "gtk/gtksignal.h", "A9" => "gdk/gdkdefs.h" );
%redeclare_postinclude = ("gdk/gdkdefs.h" => 1);

%strunames = ();
%functions = ();

%opt = ();
%shortopt = ( "-v" => "verbose", "-q" => "quiet", "-m" => "methods", "-fk" => "functions");
for (values %shortopt) { $opt{$_} = 0; }
@files = ();
%filenames = ();
$opt{"outsub"} = "cc";

# scanning the arguments
for (@ARGV)
{
  if (exists $shortopt{$_}) { $opt{$shortopt{$_}} = 1; next; }
  if (/^--(\w.*)=(.*)$/) { $opt{$1} = $2; next; } # options with args are longoptions
  if (/^--not?-(\w.*)$/) { $opt{$1} = 0; next; } 
  if (/^--(\w.*)$/) { $opt{$1} = 1; next; } 
  # anything else is a filename of course
  if (! -f "$_") { print STDERR "$_: file does not exist (skipped)\n"; next; }
  else 
  { # a filename it should be
    my $v = $_;
    if (/\//) { $v =~ s/ (\/) ([^\/]*) $/ $1.$opt{"outsub"}."\/".$2 /sex; }
    else { $v = $opt{"outsub"}."\/".$v; }
    print "[<$_>:<$v>]" if $opt{"filenames"};
    if (exists $filenames{$v}) { print STDERR "$_: its output $v is also an input file (skipped)\n"; next; }
    push @files, $_;     
    $filenames{$_} = $v;
  }
}

$addE = "";
$used_inherits = 0;

for $F (@files)
{
  open F, "<$F" or print STDERR "$F: $!" and next; $T = join( "",<F>); close F;
  
  %strunames = ();
  %functions = ();
  $addE = "";
  $used_inherits = 0;
  
  $T =~ s/(^)(\s*\#\s*define.*)$/$1$2###/mg;

  $T =~ s/ (struct\s+)(\w+) (\s+{) 
      	    (\s+ (?:\/\* [^\/]+ \*\/ \s+)? )
            (\w+)([\s\*]+)(\w+)(\;) 
         / &register_structhead($1,$2,$3,$4,$5,$6,$7,$8) /gsex;
         
  
  $T =~ s/([\w\s\*]+) (\b\w+) (\s*\(\s*) ([\w\s\*\,]+) (\s*\)\s*) (\;)
         / &register_function($1,$2,$3,$4,$5,$6) /gsex;
         
  $redeclare_preincludes = "\n";
  $redeclare_postincludes = "\n";
  for my $i (keys %strunames) { 
      if (exists $redeclare{$i}) {
          redeclare_functions($i);
          my $f = $redeclare_source{$redeclare{$i}};
          if (defined $f) {
              if (exists $redeclare_postinclude{$f})
              { 
                if (exists $filenames{$f}) { $f = $filenames{$f}; }
                $redeclare_postincludes .= "#include <$f>\n";
              }else{
                if (exists $filenames{$f}) { $f = $filenames{$f}; }
                $redeclare_preincludes .= "#include <$f>\n";
              }
          }
      }
  }
         
  print $F,": " unless $opt{"quiet"};        
  print join (" ", keys %strunames) unless $opt{"quiet"};
  print " ... " unless $opt{"quiet"};
  print join (" ", keys %functions) if $opt{"verbose"};
  
  for my $stru (keys %strunames)
  {
      my $struname = $strunames{$stru};
      
      $T =~ s/ (struct\s+ $struname \s* 
               (?:__inherits1\s*\(\s*\w+\s*\)\s*)?
               \{ [^\}]+) (\}\s*\;)
             / &add_inlines($stru,$1,$2) /gsex;
  }

  $T =~ s/###$//mg;

  $T =~ s/ (^\s*\#\s*include\s*[\<\"]) ([^<>\"]+) ([\"\>])/ &replace_include($1,$2,$3) /gmex;
  
  if ($T =~ m/define\s*GTK_SIGNAL_FUNC/) 
      { $T =~ s/ (define\s*GTK_SIGNAL_FUNC.*)$ / $1."\n#define GTK_SIG (GtkSignalFunc)\n" /mex; }

  $inherits_def =
         "\n#ifndef __inherits1"
        ."\n#ifndef __cplusplus"
        ."\n#define __inherits1(x) "
        ."\n#define __inherits2(x,y) x y ;"
        ."\n#else"
        ."\n#define __inherits1(x) : x "
        ."\n#define __inherits2(x,y)"
        ."\n#endif"
        ."\n#endif";
  $inherits_def = "" if ! $used_inherits;
  
  $T =~ s/ (\#ifdef\s*__cplusplus\s* extern \s*\"C\"\s* { \s*\#endif (?:\s*\/\*[^\*]+\*\/)? ) 
         / $1.$redeclare_preincludes.$inherits_def."\n" /sex;
        
  $T =~ s/ (\#ifdef\s*__cplusplus\s* } ) (\s*\#endif) 
         / $redeclare_postincludes.$1." \/\* end of C linkage \*\/\n\n".$addE.$2 /sex;

  print "\n" unless $opt{"quiet"};
  
  $E = $filenames{$F}; 
  if ($E =~ m/\// ) { $_ = $E; s/\/[^\/]+$// ; mkdir($_,0777) if ! -d $_; }
  open E, ">$E" or print STDERR "$E: $!\n" and next; print E $T; close E;
}

sub register_structhead
{
  my ($struP,$stru,$brac,$bracW,$base,$baseW,$name,$nameS) = @_;
        
  $struname = $stru;
  $struname =~ s/ ^_?([A-Z]) / lc($1) /ex;
  $struname =~ s/ ([A-Z]+) / "_".lc($1) /gex;
  $struname =~ s/^_+//;
  if ($struname =~ m/_/) 
  {   $strunames{"$struname"} = $stru;
  }


  if (exists $non_aggregate{"$base"} or $baseW !~ m/^\s+$/
        or $base =~ m/^[a-z]+[1-9]*$/ 
        or $base =~ m/Type$/ or $base =~ m/_type$/
        or $base =~ m/Function$/ or $base =~ m/_function$/
        or exists $non_basename{$name} )
        { return $struP.$stru.$brac.$bracW.$base.$baseW.$name.$nameS; }

  $used_inherits = 1;

  return $struP.$stru
       ."\n  __inherits1( ".$base." ) "
       .$brac." __inherits2( ".$base.", ".$name." )".$bracW ;
}

sub register_function
{
   my ($rtype,$name,$lbrc,$args,$rbrc,$semi) = @_;
   my $line = join ("",@_);
   
   print "f:$name " if $opt{"functions"};
   
   $functions{"$name"} = $rtype.$name.$lbrc.$args.$rbrc;

   return $line;
}

sub add_inlines
{
   my ($stru,$pre,$post) = @_;
   my $add = "";

   print "  [INLINES for $stru: " if $opt{"verbose"};
   
   for $f (sort keys %functions)
   {
       if ($f =~ m/^${stru}_(.*)/ 
       or defined $redeclare{$stru} and $f =~ m/^$redeclare{$stru}_(.*)/)
       {
           print $1," " if $opt{"verbose"};
           $shortname = $1;
           $longname = $f;
           $struct_name = "struct $strunames{$stru}";
           $funcdef  = $functions{$f};
           $rettype  = $funcdef;
           $rettype  =~ s/ ^ ([^\(]+) \(.* / $1 /sx;
           $rettype =~ s/\s+/ /gs;
           $rettype  =~ s/\s$//; $rettype =~ s/\w+$//;
           $rettype  =~ s/^\s//; $rettype =~ s/\s+$//;   
           $funcdecl = $funcdef;
           $funcdecl =~ s/ ^ [^\(]+ \( / \( /x;
           $funcargs = $funcdecl;
           $funcargs =~ s/ ([\w\s\*]+) (\s|\*)(\w+\s*) ([\)\,]) / $3$4 /gsx;
           $funcargs =~ s/void//g;
           $funcdecl =~ s/\s+/ /g; $funcdecl =~ s/^\s//;  $funcdecl =~ s/\s$//;
           $funcargs =~ s/\s+/ /g; $funcargs  =~ s/^\s//; $funcargs  =~ s/\s$//;
           
           $firsttype = $funcdecl;
           $firsttype =~ s/\s*\(\s*(\w+)(\s|\*).*/$1/;
           
           # the GTK people are severely misguided on returning Widget from any new-function
           $retconv = "";
           if ($shortname =~ m/^new$/ or $shortname =~ m/^new_/) {
               $rettype = "$strunames{$stru}*";
               $retconv = "($rettype)";
           }


               $add .= "  inline static $rettype _${shortname} $funcdecl; \n";
           if ($rettype eq "void")
           {
               $addE .= "inline $rettype\n"
                       ."$strunames{$stru}::_${shortname} $funcdecl \n"
                       ."                { $longname$funcargs; } \n";
           }else{
               $addE .= "inline $rettype\n"
                       ."$strunames{$stru}::_${shortname} $funcdecl \n"
                       ."                { return $retconv$longname$funcargs; } \n";
           }
             
           if ($strunames{$stru} eq "_$firsttype" || $struct_name eq $firsttype)
           {
                $funcdecl =~ s/ (\() [^\,\)]+ ([\,\)]) /$1.($2 eq ")"?")":"")/sex;
                $funcargs =~ s/ (\() [^\,\)]+ ([\,\)]) /$1."this".$2/sex;
                if ($rettype eq "void")
                {
                   # compute the type and name of the second(!) argument
                   if ( $funcargs =~ m/,/)
                   {
                      $funcdecl =~ m/ \s*\( ([^\,\)]+) [\,\)] /sx;
                      $arg2 = $1; $arg2 =~ m/^(.*[\s\*])(\w+)\s*$/s;
                      $rettype = $1; $retconv = "($2)"; $rettype =~ s/^\s+//;
                $addE .= "inline $rettype\n"
                        ."$strunames{$stru}:: ${shortname}_ $funcdecl \n"
                        ."                { $longname$funcargs; return $retconv; } \n";
                   }else{

                $addE .= "inline $rettype\n"
                        ."$strunames{$stru}:: ${shortname}_ $funcdecl \n"
                        ."                { $longname$funcargs; } \n";
                   }
                }else{
                $addE .= "inline $rettype\n"
                        ."$strunames{$stru}:: ${shortname}_ $funcdecl \n"
                        ."                { return $retconv$longname$funcargs; } \n";
                }
                $add .= "  inline        $rettype  ${shortname}_ $funcdecl; \n";
                print $shortname." " if $opt{"methods"};
           }
       }
   }
   print "]\n" if $opt{"verbose"};
   
   if (length $add > 1)
   { 
       return  $pre."\n# if defined __cplusplus || defined __cplus\n"
              .$add."\n# endif /* __cplus(plus) */\n"
              .$post;
   }else{
       return $pre.$add.$post;
   }
}

sub redeclare_functions
{
  my $stru = $_[0];

  print "[$stru -> $redeclare{$stru} -> $redeclare_source{$redeclare{$stru}}]\n" if $opt{"verbose"};
  return if not exists $redeclare{$stru} or not exists $redeclare_source{$redeclare{$stru}};
  
  my $R = $redeclare_source{$redeclare{$stru}};

  my $S = "";
  open R,"<$R" or return; $S = join("",<R>); close R;
  
  $S =~
        s/([\w\s\*]+) (\b\w+) (\s*\(\s*) ([\w\s\*\,]+) (\s*\)\s*) (\;)
         / &register_function($1,$2,$3,$4,$5,$6) /gsex;  

  return;
}

sub replace_include
{
   my ($pre,$file,$post) = @_;
   
   print $file if defined $opt{"debug-include-replace"};
   if (exists $filenames{$file})
   {
       return $pre.$filenames{$file}.$post
   }else{
       return $pre.$file.$post;
   }
}

